# Copyright (c) 2019 Matthew G. Monteleone
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from isodate import Duration, duration_isoformat, parse_datetime
from isodate.isoerror import ISO8601Error
from typing import Iterator
import logging
from enum import Enum
from datetime import datetime

logger = logging.getLogger('atlasapi.lib')


class AtlasLogNames(Enum):
    """
    The name of the log file that you want to retrieve:

    """
    MONGODB = "mongodb.gz"
    MONGOS = "mongos.gz"
    MONGOD_AUDIT = "mongodb-audit-log.gz"
    MONGOS_AUDIT = "mongos-audit-log.gz"


class LogLine(object):
    def __init__(self, raw_line):
        self.raw_line = raw_line
        try:
            raw_line_data = self.raw_line.rstrip().split(maxsplit=4)
            self.date: datetime = parse_datetime(raw_line_data[0])
            self.level: str = raw_line_data[1]
            self.facility: str = raw_line_data[2]
            self.user: str = raw_line_data[3].replace('[', '').replace(']', '')
            self.line: str = raw_line_data[-1]
            self.type: str = "Full"
        except IndexError:
            raw_line_data = raw_line.rstrip().split(maxsplit=1)
            self.date: datetime = parse_datetime(raw_line_data[0])
            self.line: str = raw_line_data[-1]
            self.type: str = "SHORT"
        except ISO8601Error:
            logger.error(f'Error Parsing line: {raw_line}')
            pass


class AtlasUnits(Enum):
    SCALAR_PER_SECOND = 'SCALAR_PER_SECOND'
    SCALAR = 'SCALAR'
    PERCENT = 'PERCENT'
    MILLISECONDS = 'MILLISECONDS'
    BYTES = 'BYTES'
    GIGABYTES = 'GIGABYTES'
    BYTES_PER_SECOND = 'BYTES_PER_SECOND'
    MEGABYTES_PER_SECOND = 'MEGABYTES_PER_SECOND'
    GIGABYTES_PER_HOUR = 'GIGABYTES_PER_HOUR'


class AtlasGranularities(object):
    """Helper class to create ISO 8601 durations to pass to the API

    To add more possible granularities, add them here.

    """
    TEN_SECOND = duration_isoformat(Duration(seconds=10))
    MINUTE = duration_isoformat(Duration(minutes=1))
    FIVE_MINUTE = duration_isoformat(Duration(minutes=5))
    HOUR = duration_isoformat(Duration(hours=1))
    DAY = duration_isoformat(Duration(days=1))



class AtlasPeriods(object):
    """Helper class to create ISO 8601 durations to send to the Atlas period parameter.

    To add more periods, add them here.
    """
    MINUTES_15 = duration_isoformat(Duration(minutes=15))
    HOURS_1 = duration_isoformat(Duration(hours=1))
    HOURS_8 = duration_isoformat(Duration(hours=8))
    HOURS_24 = duration_isoformat(Duration(hours=24))
    HOURS_48 = duration_isoformat(Duration(hours=48))
    WEEKS_1 = duration_isoformat(Duration(weeks=1))
    WEEKS_4 = duration_isoformat(Duration(weeks=4))
    MONTHS_1 = duration_isoformat(Duration(months=1))
    MONTHS_2 = duration_isoformat(Duration(months=2))
    YEARS_1 = duration_isoformat(Duration(years=1))
    YEARS_2 = duration_isoformat(Duration(years=2))


# noinspection PyCallByClass
class _GetAll(object):
    is_leaf = False

    @classmethod
    def get_all(cls) -> Iterator[str]:
        out = cls.__dict__
        for item in out:
            if '_' not in item and not item[0].isupper():
                yield cls.__getattribute__(cls, item)
            elif '_' not in item and item[0].isupper():
                sub_out = cls.__getattribute__(cls, item).__dict__
                for sub_item in sub_out:
                    if '_' not in sub_item and not sub_item[0].isupper():
                        yield cls.__getattribute__(cls, item).__dict__.get(sub_item)
                    if '_' not in sub_item and sub_item[0].isupper():
                        sub_sub_out = cls.__getattribute__(cls, item).__dict__.get(sub_item).__dict__
                        for sub_sub_item in sub_sub_out:
                            if '_' not in sub_sub_item and not sub_sub_item[0].isupper():
                                yield sub_sub_out.get(sub_sub_item)


class _GetAllLeaf(_GetAll):
    is_leaf = True


class ProviderName(Enum):
    AWS = 'Amazon Web Services'
    GCP = 'Google Cloud Platform'
    AZURE = 'Microsoft Azure'
    TENANT = 'Shared Tier'


class MongoDBMajorVersion(Enum):
    v3_4 = '3.4'
    v3_6 = '3.6'
    v4_0 = '4.0'
    v4_2 = '4.2'
    v4_4 = '4.4'
    v5_0 = '5.0'
    vX_x = 'Unknown'


class ClusterType(Enum):
    """
    The types of clusteres available in Atlas.

    GEOSHARDED is a Global write cluster sharded by geo information.

    """
    REPLICASET = 'Replica Set'
    SHARDED = 'Sharded Cluster'
    SHARDEDCLUSTER = 'Sharded Cluster'
    GEOSHARDED = 'Global Cluster'


# Copyright (c) 2022 Matthew G. Monteleone
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Atlas module

Core module which provides access to MongoDB Atlas Cloud Provider APIs
"""
from atlasapi.settings import Settings
from atlasapi.network import Network
from atlasapi.errors import *

from datetime import datetime, timezone
from dateutil.relativedelta import relativedelta
from atlasapi.specs import Host, ListOfHosts, DatabaseUsersUpdatePermissionsSpecs, DatabaseUsersPermissionsSpecs, \
    ReplicaSetTypes
from atlasapi.measurements import AtlasMeasurementTypes, AtlasMeasurementValue, AtlasMeasurement, OptionalAtlasMeasurement
from typing import Union, Iterator, List, Optional
from atlasapi.atlas_types import OptionalInt, OptionalBool, ListofDict
from atlasapi.clusters import ClusterConfig, ShardedClusterConfig, AtlasBasicReplicaSet, \
    InstanceSizeName, AdvancedOptions, TLSProtocols
from atlasapi.events import atlas_event_factory, ListOfEvents
import logging
from typing import Union, Iterable, Set, BinaryIO, Generator, Iterator
from atlasapi.errors import ErrAtlasUnauthorized, ErrAtlasBadRequest
from atlasapi.alerts import Alert
from time import time
from atlasapi.whitelist import WhitelistEntry
from atlasapi.maintenance_window import MaintenanceWindow, Weekdays
from atlasapi.lib import AtlasLogNames, LogLine, ProviderName, MongoDBMajorVersion, AtlasPeriods, AtlasGranularities, \
    AtlasUnits
from atlasapi.cloud_backup import CloudBackupSnapshot, CloudBackupRequest, SnapshotRestore, SnapshotRestoreResponse, \
    DeliveryType
from atlasapi.projects import Project, ProjectSettings
from atlasapi.teams import Team, TeamRoles
from atlasapi.atlas_users import AtlasUser
from atlasapi.organizations import Organization
from requests import get
from requests.auth import HTTPBasicAuth, HTTPDigestAuth
import gzip

logger = logging.getLogger('Atlas')


# noinspection PyProtectedMember
class Atlas:
    """Atlas constructor

    Args:
        user (str): Atlas user
        password (str): Atlas password
        group (str): Atlas group
        auth_method (Union[HTTPBasicAuth,HTTPDigestAuth]) : Authentication method to use, defaults to digest, but you
        can override to Basic if needed for use with a Proxy.
    """

    def __init__(self, user: str, password: str, group: str = None,
                 auth_method: Union[HTTPBasicAuth, HTTPDigestAuth] = HTTPDigestAuth):
        self.group = group

        # Network calls which will handled user/password for auth
        self.network = Network(user, password, auth_method)
        # APIs
        self.DatabaseUsers = Atlas._DatabaseUsers(self)
        self.Clusters = Atlas._Clusters(self)
        self.Hosts = Atlas._Hosts(self)
        self.Events = Atlas._Events(self)
        self.logger = logging.getLogger(name='Atlas')
        self.Alerts = Atlas._Alerts(self)
        self.Whitelist = Atlas._Whitelist(self)
        self.MaintenanceWindows = Atlas._MaintenanceWindows(self)
        self.CloudBackups = Atlas._CloudBackups(self)
        self.Projects = Atlas._Projects(self)
        self.Organizations = Atlas._Organizations(self)
        if not self.group:
            self.logger.warning("Note! The Atlas client has been initialized without a Group/Project, some endpoints"
                                "will not function without a Group or project.")

    class _Clusters:
        """Clusters API

        see: https://docs.atlas.mongodb.com/reference/api/clusters/

        Constructor

        Args:
            atlas (Atlas): Atlas instance
        """

        def __init__(self, atlas):
            self.atlas = atlas

        def is_existing_cluster(self, cluster) -> bool:
            """Check if the cluster exists

            Not part of Atlas api but provided to simplify some code

            Args:
                cluster (str): The cluster name

            Returns:
                bool: The cluster exists or not
            """

            try:
                self.get_single_cluster(cluster)
                return True
            except ErrAtlasNotFound:
                return False

        def get_all_clusters(self, pageNum=Settings.pageNum, itemsPerPage=Settings.itemsPerPage, iterable=False):
            """Get All Clusters

            url: https://docs.atlas.mongodb.com/reference/api/clusters-get-all/

            Keyword Args:
                pageNum (int): Page number
                itemsPerPage (int): Number of Users per Page
                iterable (bool): To return an iterable high level object instead of a low level API response

            Returns:
                AtlasPagination or dict: Iterable object representing this function OR Response payload

            Raises:
                ErrPaginationLimits: Out of limits
            """

            # Check limits and raise an Exception if needed
            ErrPaginationLimits.checkAndRaise(pageNum, itemsPerPage)

            if iterable:
                return ClustersGetAll(self.atlas, pageNum, itemsPerPage)

            uri = Settings.api_resources["Clusters"]["Get All Clusters"] % (self.atlas.group, pageNum, itemsPerPage)
            return self.atlas.network.get(Settings.BASE_URL + uri)

        def get_single_cluster(self, cluster: str) -> dict:
            """Get a Single Cluster

            url: https://docs.atlas.mongodb.com/reference/api/clusters-get-one/

            Args:
                cluster (str): The cluster name

            Returns:
                dict: Response payload
            """
            uri = Settings.api_resources["Clusters"]["Get a Single Cluster"] % (self.atlas.group, cluster)
            cluster_data = self.atlas.network.get(Settings.BASE_URL + uri)
            return cluster_data

        def get_single_cluster_advanced_options(self, cluster: str, as_obj: bool = True) -> Union[dict,
                                                                                                  AdvancedOptions]:
            """
            Retrieves advanced options from a cluster, either as a obj, or optionally as a dict.

            GET /groups/{GROUP-ID}/clusters/{CLUSTER-NAME}/processArgs

            :param cluster:
            :param as_obj: True to return, AdvancedOptions, false for a dict
            :return: AdvancedOptions object or dict
            """
            uri = Settings.api_resources["Clusters"]["Advanced Configuration Options"].format(GROUP_ID=self.atlas.group,
                                                                                              CLUSTER_NAME=cluster)
            advanced_options = self.atlas.network.get(Settings.BASE_URL + uri)

            if as_obj is True:
                return_obj = AdvancedOptions.fill_from_dict(data_dict=advanced_options)
            else:
                return_obj = advanced_options

            return return_obj

        def get_single_cluster_as_obj(self, cluster) -> Union[ClusterConfig, ShardedClusterConfig]:
            """Get a Single Cluster as data

            url: https://docs.atlas.mongodb.com/reference/api/clusters-get-one/

            Args:
                cluster (str): The cluster name

            Returns:
                ClusterConfig: Response payload
            """
            cluster_data = self.get_single_cluster(cluster=cluster)
            try:
                if cluster_data.get('clusterType', None) == 'SHARDED':
                    logger.info("Cluster Type is SHARDED, Returning a ShardedClusterConfig")
                    out_obj = ShardedClusterConfig.fill_from_dict(data_dict=cluster_data)
                elif cluster_data.get('clusterType', None) == 'REPLICASET':
                    logger.info("Cluster Type is REPLICASET, Returning a ClusterConfig")
                    out_obj = ClusterConfig.fill_from_dict(data_dict=cluster_data)
                else:
                    logger.info("Cluster Type is not recognized, Returning a REPLICASET")
                    out_obj = ClusterConfig.fill_from_dict(data_dict=cluster_data)
            except Exception as e:
                raise e
            return out_obj

        def create_cluster(self, cluster: ClusterConfig) -> dict:
            """Create a cluster

            url: POST /api/atlas/v1.0/groups/{GROUP-ID}/clusters

            Args:
                cluster (ClusterConfig): A Cluster Config Object

            Returns:
                dict: Response payload
            """
            uri = Settings.api_resources["Clusters"]["Create a Cluster"].format(GROUP_ID=self.atlas.group)
            logger.info("Initiating Call to Atlas API at {}".format(Settings.BASE_URL + uri))
            logger.info("Cluster Config = {}".format(cluster))

            cluster_data = self.atlas.network.post(Settings.BASE_URL + uri, payload=cluster.as_create_dict())
            return cluster_data

        def create_basic_rs(self, name: str,
                            size: InstanceSizeName = InstanceSizeName.M10,
                            disk_size: int = 10,
                            provider: ProviderName = ProviderName.AWS,
                            region: str = 'US_WEST_2',
                            version: MongoDBMajorVersion = MongoDBMajorVersion.v4_0) -> AtlasBasicReplicaSet:
            """
            Simplified method for creating a basic replica set with basic options.

            :param name: The name for the cluster
            :param size: The Atlas Instance size, found in The InstanceSizeName enum
            :param disk_size: The size in GB for disk
            :param provider: The cloud provider, found in ProviderName enum
            :param region: The provider region to place the cluster.
            :param version: The mongodb major version (enum)
            :return: AtlasBasicReplicaSet
            """
            cluster: AtlasBasicReplicaSet = AtlasBasicReplicaSet(name=name, size=size, disk_size=disk_size,
                                                                 provider=provider, region=region,
                                                                 version=version)
            result = self.create_cluster(cluster=cluster.config)

            cluster.config_running = result
            return cluster

        def delete_cluster(self, cluster: str, areYouSure: bool = False):
            """Delete a Cluster

            url: https://docs.atlas.mongodb.com/reference/api/clusters-delete-one/

            Args:
                cluster (str): Cluster name

            Keyword Args:
                areYouSure (bool): safe flag to don't delete a cluster by mistake

            Returns:
                dict: Response payload

            Raises:
                ErrConfirmationRequested: Need a confirmation to delete the cluster
                :type areYouSure: bool
                :param cluster: Cluster name
                :param areYouSure: safe flag to don't delete a cluster by mistake
            """
            if areYouSure:
                uri = Settings.api_resources["Clusters"]["Delete a Cluster"] % (self.atlas.group, cluster)
                return self.atlas.network.delete(Settings.BASE_URL + uri)
            else:
                raise ErrConfirmationRequested(
                    "Please set areYouSure=True on delete_cluster call if you really want to delete [%s]" % cluster)

        def modify_cluster(self, cluster: str, cluster_config: Union[ClusterConfig, dict]) -> dict:
            """Modify a Cluster

            Modifies an existing cluster in the project. Either from a full ClusterConfig object, or from a simple
            dict which contains the elements desired.


            url: https://docs.atlas.mongodb.com/reference/api/clusters-modify-one/


            :param cluster: The name of the cluster to modify
            :param cluster_config: A ClusterConfig object containing the new configuration,
                                   or a dict containing fragment.
            :return: dict:  A dictionary of the new cluster config
            """
            uri = Settings.api_resources["Clusters"]["Modify a Cluster"].format(GROUP_ID=self.atlas.group,
                                                                                CLUSTER_NAME=cluster)
            try:
                self.get_single_cluster_as_obj(cluster=cluster)
            except ErrAtlasNotFound:
                logger.error('Could not find existing cluster {}'.format(cluster))
                raise ValueError('Could not find existing cluster {}'.format(cluster))

            if type(cluster_config) in (ShardedClusterConfig, ClusterConfig):
                logger.debug("We received a full cluster_config, converting to dict")
                try:
                    new_config = cluster_config.as_modify_dict()
                except Exception as e:
                    logger.error('Error while trying to parse the new configuration')
                    raise e
            else:
                logger.debug("We received a simple dict for cluster config, sending without converting.")
                new_config = cluster_config
            value_returned = self.atlas.network.patch(uri=Settings.BASE_URL + uri, payload=new_config)
            return value_returned

        def modify_cluster_instance_size(self, cluster: str, new_cluster_size: InstanceSizeName) -> dict:
            """
            Modifies existing cluster by changing only the instance size.

            Helper function using modify_cluster
            :param cluster: The cluster name
            :param new_cluster_size: InstanceSizeName: The new size to use.
            :return: dict: the new cluster configuration dict

            """
            # Check new_cluster_size type:
            try:
                assert isinstance(new_cluster_size, InstanceSizeName)
            except AssertionError:
                raise TypeError(
                    f'new_cluster_size must be an instance of InstanceSizeName, not a {type(new_cluster_size)}')

            # First check to see if this is a new size, and if the cluster exists
            try:
                existing_config = self.get_single_cluster_as_obj(cluster=cluster)
                if existing_config.providerSettings.instance_size_name == new_cluster_size:
                    logger.error("New size is the same as old size.")
                    raise ValueError("New size is the same as old size.")
            except ErrAtlasNotFound:
                logger.error('Could not find existing cluster {}'.format(cluster))
                raise ValueError('Could not find existing cluster {}'.format(cluster))

            existing_config.providerSettings.instance_size_name = new_cluster_size
            return self.modify_cluster(cluster=cluster, cluster_config=existing_config)

        def modify_cluster_advanced_options(self, cluster: str,
                                            advanced_options: AdvancedOptions,
                                            as_obj: bool = True) -> Union[AdvancedOptions, dict]:
            """
            Modifies cluster advanced options using a AdvancedOptions object.

            PATCH /groups/{GROUP-ID}/clusters/{CLUSTER-NAME}/processArgs

            :param cluster: The clutster name
            :param advanced_options: An AdvancedOptions object with the options to be set.
            :param as_obj: Return the new AdvancedOptions as an object.
            :return:
            """
            uri = Settings.api_resources["Clusters"]["Advanced Configuration Options"].format(GROUP_ID=self.atlas.group,
                                                                                              CLUSTER_NAME=cluster)

            value_returned = self.atlas.network.patch(uri=Settings.BASE_URL + uri, payload=advanced_options.as_dict)

            if as_obj is True:
                return_obj = AdvancedOptions.fill_from_dict(data_dict=value_returned)
            else:
                return_obj = value_returned

            return return_obj

        def modify_cluster_tls(self, cluster: str,
                               TLS_protocol: TLSProtocols,
                               as_obj: bool = True) -> TLSProtocols:
            """
            Modifies cluster TLS settings.


            """
            uri = Settings.api_resources["Clusters"]["Advanced Configuration Options"].format(GROUP_ID=self.atlas.group,
                                                                                              CLUSTER_NAME=cluster)

            advanced_options = AdvancedOptions(minimumEnabledTlsProtocol=TLS_protocol)

            value_returned = self.atlas.network.patch(uri=Settings.BASE_URL + uri, payload=advanced_options.as_dict)

            if as_obj is True:
                return_obj = AdvancedOptions.fill_from_dict(data_dict=value_returned)
            else:
                return_obj = value_returned

            return return_obj

        def pause_cluster(self, cluster: str, toggle_if_paused: bool = False) -> dict:
            """
            Pauses/Unpauses a cluster.

            If you wish to unpause, set the toggle_if_paused param to True.
            :rtype: dict
            :param cluster: The name of the cluster
            :param toggle_if_paused: Set to true to unpause a paused clsuter.
            :return: dict: The updated config
            """
            existing_config = self.get_single_cluster_as_obj(cluster=cluster)
            logger.info('The cluster state is currently Paused= {}'.format(existing_config.paused))
            if existing_config.paused is True and toggle_if_paused is False:
                logger.error("The cluster is already paused. Use unpause instead.")
                raise ErrAtlasBadRequest(400,
                                         {'msg': 'The cluster is already paused. Use toggle_if_paused to unpause.'})
            elif existing_config.paused is True and toggle_if_paused is True:
                logger.warning('Cluster is paused, will toggle to unpaused, since toggle_if paused is true')
                new_config = dict(paused=False)
            else:
                new_config = dict(paused=True)
            return self.modify_cluster(cluster=cluster, cluster_config=new_config)

        def test_failover(self, cluster: str) -> Optional[dict]:
            """
            Triggers a primary failover for a cluster

            Used for testing cluster resiliency.

            :rtype: dict
            :param cluster:
            :return: And empty dict
            """
            uri = Settings.api_resources["Clusters"]["Test Failover"].format(GROUP_ID=self.atlas.group,
                                                                             CLUSTER_NAME=cluster)

            return self.atlas.network.post(uri=Settings.BASE_URL + uri, payload={})

    class _Hosts:
        """Hosts API

        see: https://docs.atlas.mongodb.com/reference/api/monitoring-and-logs/#monitoring-and-logs

        Constructor

        Args:
            atlas (Atlas): Atlas instance
        """

        def __init__(self, atlas):
            self.atlas: atlas = atlas
            self.logger = logging.getLogger('Atlas.Hosts')
            self.host_list_with_measurements: Optional[List[Host]] = list()
            self.host_list: Optional[List[Host]] = list()

        def _get_all_hosts(self, pageNum=Settings.pageNum,
                           itemsPerPage=Settings.itemsPerPage,
                           iterable=False):
            """Get All Hosts (actually processes)

            Internal use only, actual data retrieval comes from properties host_list and host_names
            url: https://docs.atlas.mongodb.com/reference/api/alerts-get-all-alerts/

            Keyword Args:
                pageNum (int): Page number
                itemsPerPage (int): Number of Users per Page
                iterable (bool): To return an iterable high level object instead of a low level API response

            Returns:
                ListOfHosts or dict: Iterable object representing this function OR Response payload

            Raises:
                ErrPaginationLimits: Out of limits
                :rtype: Union[ListOfHosts, dict]
                :type iterable: OptionalBool
                :type itemsPerPage: OptionalInt
                :type pageNum: OptionalInt

            """

            # Check limits and raise an Exception if needed
            ErrPaginationLimits.checkAndRaise(pageNum, itemsPerPage)

            if iterable:
                item_list = list(HostsGetAll(self.atlas, pageNum, itemsPerPage))
                obj_list = list()
                for item in item_list:
                    obj_list.append(Host(item))

                return_val = obj_list
            else:
                uri = Settings.api_resources["Monitoring and Logs"]["Get all processes for group"].format(
                    group_id=self.atlas.group,
                    page_num=pageNum,
                    items_per_page=itemsPerPage)

                return_val = self.atlas.network.get(Settings.BASE_URL + uri)

            return return_val

        def fill_host_list(self, for_cluster: Optional[str] = None) -> List[Host]:
            """
            Fills the `self.hostname` property with the current hosts for the project/group.

            Optionally, one can specify the `for_cluster` parameter to fill the host list with
            hosts only from the specified cluster.

            Args:
                for_cluster (str): The name of the cluster for filter the host list.

            Returns:
                List[Host]: A lost of  `Host` objects
            """
            host_list = self._get_all_hosts(iterable=True)
            if for_cluster:
                out_list = list()
                for host in host_list:
                    if host.cluster_name.lower() == for_cluster.lower():
                        out_list.append(host)
                    elif host.cluster_name_alias.lower() == for_cluster.lower():
                        out_list.append(host)
                self.host_list = out_list
            else:
                self.host_list = self._get_all_hosts(iterable=True)

            return self.host_list

        @property
        def host_names(self) -> Iterable[str]:
            """Returns a simple list of host names without port

            :rtype: Iterator[str]
            """

            for host in self.host_list:
                yield host.hostname

        @property
        def cluster_list(self) -> Set[str]:
            """
            Returns a list of clusters found in the hosts for this group.

            Returns:
                Set[str}: A set of cluster names
            """
            cluster_list = set()
            for host in self.host_list:
                cluster_list.add(host.cluster_name)
            return cluster_list

        def host_list_by_cluster(self, cluster_name: str) -> Iterable[Host]:
            """
            Returns hosts belonging to the named cluster.
            Args:
                cluster_name (str):
            Returns:
                 Iterable[Host]: An interator of Host Objects.
            """
            for host in self.host_list:
                if host.cluster_name == cluster_name:
                    yield host

        def update_host_list(self, host_obj: Host) -> None:
            """
            Places a host into the host_list property.

            Args:
                host_obj: Host: A host object with measurements.

            Returns:
                None:
            """
            for n, i in enumerate(self.host_list):
                if i == host_obj:
                    self.host_list[n] = host_obj
            self.logger.info("Added new host item")

        @property
        def host_list_primaries(self) -> Iterable[Host]:
            """Yields only hosts which are currently primary.

            """
            for each_host in self.host_list:
                if each_host.type == ReplicaSetTypes.REPLICA_PRIMARY:
                    yield each_host

        @property
        def host_list_secondaries(self) -> Iterable[Host]:
            """Yields only hosts which are currently secondaries.

            """
            for each_host in self.host_list:
                if each_host.type == ReplicaSetTypes.REPLICA_SECONDARY:
                    yield each_host

        def get_measurement_for_hosts(self, granularity: Optional[AtlasGranularities] = None,
                                      period: Optional[AtlasPeriods] = None,
                                      measurement: Optional[AtlasMeasurementTypes] = None, return_data: bool = False):
            """Get measurement(s) for all hosts in the host_list


                        Populates all hosts in the host_list with the requested metric.

                        Multiple calls will append additional metrics to the same host object.

                        Please note that using the `return_data`  param will also return the updated
                        host objects, which may unnecessarily consume memory.

                        Keyword Args:
                            granularity (AtlasGranularities): the desired granularity
                            period (AtlasPeriods): The desired period
                            measurement (AtlasMeasurementTypes) : The desired measurement or Measurement class



                            :param return_data:
                            :rtype: List[measurements.AtlasMeasurement]
                            :type period: AtlasPeriods
                            :type granularity: AtlasGranularities
                            :type measurement: measurements.AtlasMeasurementTypes
                        """

            if not self.host_list:
                raise (ValueError('Before retrieving measurements for hosts, you must retrieve the host list'
                                  ' by using one of the `fill_host_list`.'))
            return_list = list()
            for each_host in self.host_list:
                logger.info('Pulling measurements for {} hosts'.format(len(self.host_list)))
                logger.info(f'Measurement: {measurement}')
                logger.info('Metric: {}'.format(granularity.__str__()))
                logger.info('For Period: {}'.format(period.__str__()))
                try:
                    logger.debug(f'The data type of measurement is is {type(measurement)}')
                    returned_data = self._get_measurement_for_host(each_host, granularity=granularity,
                                                                   period=period,
                                                                   measurement=measurement)
                    each_host.measurements = list(returned_data)
                except Exception as e:
                    logger.error('An error occurred while retrieving metrics for host: {}.'
                                 'The error was {}'.format(each_host.hostname, e))
                    logger.warning('Will continue with next host. . . ')
                return_list.append(each_host)
            self.host_list_with_measurements = return_list
            if return_data:
                return return_list

        def get_log_for_host(self, host_obj: Host,
                             log_name: AtlasLogNames = AtlasLogNames.MONGODB,
                             date_from: datetime = None,
                             date_to: datetime = None,
                             ) -> BinaryIO:
            """
            Retrieves the designated logfile archive of designated log_name and for the designated dates,
            and returns a binary file like object.

            Args:
                host_obj (Host): And atlas Host object to retrieve logs for
                log_name (AtlasLogNames): an AtlasLogNames type
                date_from (datetime.datetime): The datetime to start from
                date_to (datetime.datetime): The datetime to gather till

            Returns:
                BinaryIO: A BinaryIO object containing the gzipped log file.

            """
            uri = Settings.api_resources["Monitoring and Logs"]["Get the log file for a host in the cluster"].format(
                group_id=self.atlas.group,
                host=host_obj.hostname,
                logname=log_name.value,
                date_from=date_from,
                date_to=date_to
            )
            # Build the request
            if date_to is None and date_from is None:
                logger.info('No dates passed so we are not going to send date params, API default will be used.')
                uri = Settings.BASE_URL + uri
            elif date_to is None and date_from is not None:
                logger.info('Received only a date_from, so sending only startDate')
                uri = Settings.BASE_URL + uri + f'?startDate={int(round(date_from.timestamp()))}'
            elif date_to is not None and date_from is None:
                uri = Settings.BASE_URL + uri + f'?endDate={int(round(date_to.timestamp()))}'
            else:
                uri = Settings.BASE_URL + uri + f'?endDate={int(round(date_to.timestamp()))}' \
                                                f'&startDate={int(round(date_from.timestamp()))}'
            logger.info(f'The URI used will be {uri}')
            return_val = self.atlas.network.get_file(uri)
            return return_val

        def get_loglines_for_host(self, host_obj: Host,
                                  log_name: AtlasLogNames = AtlasLogNames.MONGODB,
                                  date_from: datetime = None,
                                  date_to: datetime = None,
                                  ) -> Iterable[LogLine]:
            """
            Gathers the designated log file from Atlas, and then returns the lines therein contained.

            Does so by downloading the gzip file into memory, ungzipping and then unpacking each log line
            as a LogLine Object.

            Args:
                host_obj (Host): An atlas Host object to retrive logs for
                log_name (str): an AtlasLogNames type
                date_from (datetime): The datetime to start from
                date_to (datetime): The datetime to gather till

            Returns:
                Iterable[LogLine] : Yeilds LogLine objects, one for each logline found in the file.
            """
            result = self.get_log_for_host(host_obj, log_name, date_from, date_to)
            result.seek(0)
            content = gzip.GzipFile(fileobj=result)
            for each in content.readlines():
                yield LogLine(each.decode())

        def get_logs_for_project(self,
                                 log_name: AtlasLogNames = AtlasLogNames.MONGODB,
                                 date_from: datetime = None,
                                 date_to: datetime = None) -> Iterable[Host]:
            """
            Yields A Host object per Host in the project with  a File-like objects containing the gzipped log file
            requested for each  host in the project using the same date filters and log_name (type) in the log_files
            property.

            Currently the log_file property (List) is usually with only one item.
            Args:
                log_name (AtlasLogNames): The type of log to be retrieved
                date_from (datetime) : Start of log entries
                date_to (datetime): End of log entries

            Returns:
                Iterable[Host]: Yields Host objects, with full host information as well as the logfile in the log_files
                property.
            """
            for each_host in self.host_list:
                try:
                    log_file: BinaryIO = self.get_log_for_host(host_obj=each_host,
                                                               log_name=log_name,
                                                               date_from=date_from,
                                                               date_to=date_to)
                except Exception as e:
                    raise e
                each_host.add_log_file(name=log_name, file=log_file)
                yield each_host

        def get_logs_for_cluster(self,
                                 cluster_name: str,
                                 log_name: AtlasLogNames = AtlasLogNames.MONGODB,
                                 date_from: datetime = None,
                                 date_to: datetime = None) -> Iterable[Host]:
            """
            Yields A Host object per Host in the passed cluster with  a File-like objects containing the gzipped log file
            requested for each  host in the project using the same date filters and log_name (type) in the log_files
            property.

            Currently the log_file property (List) is usually with only one item.
            Args:
                log_name (AtlasLogNames): The type of log to be retrieved
                date_from (datetime) : Start of log entries
                date_to (datetime): End of log entries

            Returns:
                Iterable[Host]: Yields Host objects, with full host information as well as the logfile in the log_files
                property.
            """
            for each_host in self.host_list_by_cluster(cluster_name=cluster_name):
                try:
                    log_file: BinaryIO = self.get_log_for_host(host_obj=each_host,
                                                               log_name=log_name,
                                                               date_from=date_from,
                                                               date_to=date_to)
                except Exception as e:
                    raise e
                each_host.add_log_file(name=log_name, file=log_file)
                yield each_host

        def _get_measurement_for_host(self, host_obj: Host,
                                      granularity: Optional[AtlasGranularities] = None,
                                      period: Optional[AtlasPeriods] = None,
                                      measurement: Optional[AtlasMeasurementTypes] = None,
                                      pageNum: int = Settings.pageNum,
                                      itemsPerPage: int = Settings.itemsPerPage,
                                      iterable: bool = True) -> Union[dict, Iterable[AtlasMeasurement]]:
            """Get  measurement(s) for a host

            Internal use only, should come from the host obj itself.

            Returns measurements for the passed Host object.

            url: https://docs.atlas.mongodb.com/reference/api/process-measurements/


            Accepts either a single measurement, but will retrieve more than one measurement
            if the measurement (using the AtlasMeasurementTypes class)

            /api/atlas/v1.0/groups/{GROUP-ID}/processes/{HOST}:{PORT}/measurements

            Keyword Args:
                host_obj (Host): the host
                granularity (AtlasGranuarities): the desired granularity
                period (AtlasPeriods): The desired period
                measurement (AtlasMeasurementTypes) : The desired measurement or Measurement class
                pageNum (int): Page number
                itemsPerPage (int): Number of Users per Page
                iterable (bool): To return an iterable high level object instead of a low level API response

            Returns:
                 Iterable[AtlasMeasurement] or dict: Iterable object representing this function OR Response payload

            Raises:
                ErrPaginationLimits: Out of limits

                :rtype: List[measurements.AtlasMeasurement]
                :type iterable: OptionalBool
                :type itemsPerPage: OptionalInt
                :type pageNum: OptionalInt
                :type period: AtlasPeriods
                :type granularity: AtlasGranularities
                :type host_obj: Host
                :type measurement: measurements.AtlasMeasurementTypes

            """

            # Check limits and raise an Exception if needed
            ErrPaginationLimits.checkAndRaise(pageNum, itemsPerPage)
            if measurement is None:
                measurement = AtlasMeasurementTypes.Cache.dirty
            if period is None:
                period = AtlasPeriods.WEEKS_1

            if granularity is None:
                granularity = AtlasGranularities.HOUR
            # Check to see if we received a leaf or branch of the measurements
            logger.debug(f'Measurement is: {measurement}')
            logger.debug(f'Measurement object type is {type(measurement)}')
            try:
                parent = super(measurement)
                self.logger.error('We received a branch, whos parent is {}'.format(parent.__str__()))
                leaves = measurement.get_all()
                measurement_list = list(leaves)
                measurement = '&m='.join(measurement_list)
            except TypeError:
                self.logger.info('We received a leaf')

            # Build the URL

            if isinstance(measurement, tuple):
                logger.debug(f'Somehow got a tuple for measurement {measurement}. Need to get the str')
                measurement: tuple = measurement[0]
            if isinstance(granularity, tuple):
                logger.info(f'Somehow got a tuple for granularity {granularity}. Need to get the str')
                granularity: tuple = granularity[0]
            if isinstance(period, tuple):
                logger.debug(f'Somehow got a tuple for period {period}. Need to get the str')
                period: tuple = period[0]

            uri = Settings.api_resources["Monitoring and Logs"]["Get measurement for host"].format(
                group_id=self.atlas.group,
                host=host_obj.hostname,
                port=host_obj.port,
                granularity=granularity,
                period=period,
                measurement=measurement
            )
            logger.debug(f'The URI used will be {uri}')
            # Build the request
            return_val = self.atlas.network.get(Settings.BASE_URL + uri)

            if iterable:
                measurements = return_val.get('measurements')
                measurements_count = len(measurements)
                self.logger.info('There are {} measurements.'.format(measurements_count))

                for each in measurements:
                    measurement_obj = AtlasMeasurement(name=each.get('name'),
                                                       period=period,
                                                       granularity=granularity)
                    for each_and_every in each.get('dataPoints'):
                        measurement_obj.measurements = AtlasMeasurementValue(each_and_every)

                yield measurement_obj

            else:
                return return_val

    class _Events:
        """Events API

        see: https://docs.atlas.mongodb.com/reference/api/events/

        Constructor

        Args:
            atlas (Atlas): Atlas instance
        """

        def __init__(self, atlas):
            self.atlas = atlas  # type: Atlas
            self.logger = logging.getLogger('Atlas.Events')  # type: logging

        def _get_all_project_events(self, since_datetime: datetime = None, pageNum: int = Settings.pageNum,
                                    itemsPerPage: int = Settings.itemsPerPage,
                                    iterable: bool = False) -> ListOfEvents:
            """Get All Project Events

            Internal use only, actual data retrieval comes from properties all
            url: https://docs.atlas.mongodb.com/reference/api/events-projects-get-all/

            Keyword Args:
                pageNum (int): Page number
                itemsPerPage (int): Number of Users per Page
                iterable (bool): To return an iterable high level object instead of a low level API response

            Returns:
                ListOfEvents or dict: Iterable object representing this function OR Response payload

            Raises:
                ErrPaginationLimits: Out of limits
                :rtype: Union[ListOfEvents, dict]
                :type iterable: OptionalBool
                :type itemsPerPage: OptionalInt
                :type pageNum: OptionalInt

            """

            # Check limits and raise an Exception if needed
            ErrPaginationLimits.checkAndRaise(pageNum, itemsPerPage)

            if iterable:
                item_list = list(EventsGetForProject(self.atlas, since_datetime, pageNum, itemsPerPage))
                obj_list: ListOfEvents = list()
                for item in item_list:
                    obj_list.append(atlas_event_factory(item))
                return obj_list

            if since_datetime:
                uri = Settings.api_resources["Events"]["Get Project Events Since Date"].format(
                    min_date=since_datetime.isoformat(),
                    group_id=self.atlas.group,
                    page_num=pageNum,
                    items_per_page=itemsPerPage)

                return_val = self.atlas.network.get(Settings.BASE_URL + uri)

            else:
                uri = Settings.api_resources["Events"]["Get All Project Events"].format(
                    group_id=self.atlas.group,
                    page_num=pageNum,
                    items_per_page=itemsPerPage)

                return_val = self.atlas.network.get(Settings.BASE_URL + uri)
            return return_val

        @property
        def all(self) -> ListOfEvents:
            """
            Returns all events for the current project/group.

            Returns:
                ListOfEvents: A list of event objects.

            """
            return self._get_all_project_events(iterable=True)

        def since(self, since_datetime: datetime) -> ListOfEvents:
            """
            Returns all events since the passed datetime. (UTC)

            Returns:
                ListOfEvents:
            """
            return self._get_all_project_events(iterable=True, since_datetime=since_datetime)

    class _DatabaseUsers:
        """Database Users API

        see: https://docs.atlas.mongodb.com/reference/api/database-users/

        Constructor

        Args:
            atlas (Atlas): Atlas instance
        """

        def __init__(self, atlas) -> None:
            self.atlas = atlas

        def get_all_database_users(self, pageNum: int = Settings.pageNum, itemsPerPage: int = Settings.itemsPerPage,
                                   iterable: bool = False):
            """Get All Database Users

            url: https://docs.atlas.mongodb.com/reference/api/database-users-get-all-users/

            Keyword Args:
                pageNum (int): Page number
                itemsPerPage (int): Number of Users per Page
                iterable (bool): To return an iterable high level object instead of a low level API response

            Returns:
                AtlasPagination or dict: Iterable object representing this function OR Response payload

            Raises:
                ErrPaginationLimits: Out of limits
            """

            # Check limits and raise an Exception if needed
            ErrPaginationLimits.checkAndRaise(pageNum, itemsPerPage)

            if iterable:
                return DatabaseUsersGetAll(self.atlas, pageNum, itemsPerPage)

            uri = Settings.api_resources["Database Users"]["Get All Database Users"] % (
                self.atlas.group, pageNum, itemsPerPage)
            return self.atlas.network.get(Settings.BASE_URL + uri)

        def get_a_single_database_user(self, user: str) -> dict:
            """Get a Database User

            url: https://docs.atlas.mongodb.com/reference/api/database-users-get-single-user/

            Args:
                user (str): User

            Returns:
                dict: Response payload
            """
            uri = Settings.api_resources["Database Users"]["Get a Single Database User"] % (self.atlas.group, user)
            return self.atlas.network.get(Settings.BASE_URL + uri)

        def create_a_database_user(self, permissions) -> dict:
            """Create a Database User

            url: https://docs.atlas.mongodb.com/reference/api/database-users-create-a-user/

            Args:
                permissions (DatabaseUsersPermissionsSpec): Permissions to apply

            Returns:
                dict: Response payload
            """
            uri = Settings.api_resources["Database Users"]["Create a Database User"] % self.atlas.group
            return self.atlas.network.post(Settings.BASE_URL + uri, permissions.getSpecs())

        def update_a_database_user(self, user: str, permissions: DatabaseUsersUpdatePermissionsSpecs) -> dict:
            """Update a Database User

            url: https://docs.atlas.mongodb.com/reference/api/database-users-update-a-user/

            Args:
                user (str): User
                permissions (DatabaseUsersUpdatePermissionsSpecs): Permissions to apply

            Returns:
                dict: Response payload
            """
            uri = Settings.api_resources["Database Users"]["Update a Database User"] % (self.atlas.group, user)
            logger.info(Settings.BASE_URL + uri, permissions.getSpecs())
            return self.atlas.network.patch(Settings.BASE_URL + uri, permissions.getSpecs())

        def delete_a_database_user(self, user: str) -> dict:
            """Delete a Database User

            url: https://docs.atlas.mongodb.com/reference/api/database-users-delete-a-user/

            Args:
                user (str): User to delete

            Returns:
                dict: Response payload
            """
            uri = Settings.api_resources["Database Users"]["Delete a Database User"] % (self.atlas.group, user)
            return self.atlas.network.delete(Settings.BASE_URL + uri)

    class _Alerts:
        """Alerts API

        see: https://docs.atlas.mongodb.com/reference/api/alerts/

        Constructor

        Args:
            atlas (Atlas): Atlas instance
        """

        def __init__(self, atlas):
            self.atlas = atlas

        def get_all_alerts(self, status=None, pageNum=Settings.pageNum, itemsPerPage=Settings.itemsPerPage,
                           iterable=False):
            """Get All Alerts

            url: https://docs.atlas.mongodb.com/reference/api/alerts-get-all-alerts/

            Keyword Args:
                status (AlertStatusSpec): filter on alerts status
                pageNum (int): Page number
                itemsPerPage (int): Number of Users per Page
                iterable (bool): To return an iterable high level object instead of a low level API response

            Returns:
                AtlasPagination or dict: Iterable object representing this function OR Response payload

            Raises:
                ErrPaginationLimits: Out of limits
            """

            # Check limits and raise an Exception if needed
            ErrPaginationLimits.checkAndRaise(pageNum, itemsPerPage)

            if iterable:
                return AlertsGetAll(self.atlas, status, pageNum, itemsPerPage)

            if status:
                uri = Settings.api_resources["Alerts"]["Get All Alerts with status"] % (
                    self.atlas.group, status, pageNum, itemsPerPage)
            else:
                uri = Settings.api_resources["Alerts"]["Get All Alerts"] % (self.atlas.group, pageNum, itemsPerPage)

            return self.atlas.network.get(Settings.BASE_URL + uri)

        def get_an_alert(self, alert: str) -> Alert:
            """Get an Alert

            url: https://docs.atlas.mongodb.com/reference/api/alerts-get-alert/

            Args:
                alert (str): The alert id

            Returns:
                dict: Response payload
            """
            uri = Settings.api_resources["Alerts"]["Get an Alert"] % (self.atlas.group, alert)
            result = self.atlas.network.get(Settings.BASE_URL + uri)
            return Alert(result)

        # TODO: FIX THIS
        def acknowledge_an_alert(self, alert, until, comment=None):
            """Acknowledge an Alert

            url: https://docs.atlas.mongodb.com/reference/api/alerts-acknowledge-alert/

            Args:
                alert (str): The alert id
                until (datetime): Acknowledge until

            Keyword Args:
                comment (str): The acknowledge comment

            Returns:
                dict: Response payload
                :param comment:
            """

            # data = {"acknowledgedUntil": until.isoformat(timespec='seconds')}
            data = {"acknowledgedUntil": "2017-10-23T08:28:35Z"}
            if comment:
                data["acknowledgementComment"] = comment

            uri = Settings.api_resources["Alerts"]["Acknowledge an Alert"] % (self.atlas.group, alert)
            return self.atlas.network.patch(Settings.BASE_URL + uri, data)

        # TODO: FIX THIS
        def unacknowledge_an_alert(self, alert):
            """Acknowledge an Alert

            url: https://docs.atlas.mongodb.com/reference/api/alerts-acknowledge-alert/

            Args:
                alert (str): The alert id

            Returns:
                dict: Response payload
            """

            # see https://docs.atlas.mongodb.com/reference/api/alerts-acknowledge-alert/#request-body-parameters
            # To unacknowledge a previously acknowledged alert, set the field value to the past.
            now = datetime.now(timezone.utc)
            until = now - relativedelta(days=1)
            return self.acknowledge_an_alert(alert, until)

        # TODO: FIX THIS
        def acknowledge_an_alert_forever(self, alert, comment=None):
            """Acknowledge an Alert forever

            url: https://docs.atlas.mongodb.com/reference/api/alerts-acknowledge-alert/

            Args:
                alert (str): The alert id

            Keyword Args:
                comment (str): The acknowledge comment

            Returns:
                dict: Response payload
            """

            # see https://docs.atlas.mongodb.com/reference/api/alerts-acknowledge-alert/#request-body-parameters
            # To acknowledge an alert “forever”, set the field value to 100 years in the future.
            now = datetime.now(timezone.utc)
            until = now + relativedelta(years=100)
            return self.acknowledge_an_alert(alert, until, comment)

    class _Whitelist:
        """Whitelist API

        see: https://docs.atlas.mongodb.com/reference/api/whitelist/

        Constructor

        Args:
            atlas (Atlas): Atlas instance
        """

        def __init__(self, atlas):
            self.atlas = atlas

        def get_all_whitelist_entries(self, pageNum: int = Settings.pageNum, itemsPerPage: int = Settings.itemsPerPage,
                                      iterable: bool = False) -> Iterable[WhitelistEntry]:
            """Get All whitelist entries

            url: https://docs.atlas.mongodb.com/reference/api/whitelist-get-all/

            Keyword Args:
                pageNum (int): Page number
                itemsPerPage (int): Number of Users per Page
                iterable (bool): To return an iterable high level object instead of a low level API response

            Returns:
                AtlasPagination or dict: Iterable object representing this function OR Response payload

            Raises:
                ErrPaginationLimits: Out of limits
            """

            # Check limits and raise an Exception if needed
            ErrPaginationLimits.checkAndRaise(pageNum, itemsPerPage)

            if iterable:
                return WhitelistGetAll(self.atlas, pageNum, itemsPerPage)

            uri = Settings.api_resources["Whitelist"]["Get All Whitelist Entries"] % (
                self.atlas.group, pageNum, itemsPerPage)

            response = self.atlas.network.get(Settings.BASE_URL + uri)

            for entry in response.get('results', []):
                yield WhitelistEntry.fill_from_dict(entry)

        def get_whitelist_entry(self, ip_address: str) -> WhitelistEntry:
            """Get a whitelist entry

            url: https://docs.atlas.mongodb.com/reference/api/whitelist-get-one-entry/

            Args:
                ip_address (str): ip address to fetch from whitelist

            Returns:
                WhitelistEntry: Response payload
            """
            uri = Settings.api_resources["Whitelist"]["Get Whitelist Entry"] % (
                self.atlas.group, ip_address)
            response: dict = self.atlas.network.get(Settings.BASE_URL + uri)
            return WhitelistEntry.fill_from_dict(response)

        def create_whitelist_entry(self, ip_address: str, comment: str) -> List[WhitelistEntry]:
            """Create a whitelist entry

            url: https://docs.atlas.mongodb.com/reference/api/whitelist-add-one/

            Args:
                ip_address (str): ip address to add to whitelist
                comment (str): comment describing the whitelist entry

            Returns:
                List[WhitelistEntry]: Response payload
            """
            uri = Settings.api_resources["Whitelist"]["Create Whitelist Entry"] % self.atlas.group

            whitelist_entry = {'ipAddress': ip_address, 'comment': comment}
            response = self.atlas.network.post(Settings.BASE_URL + uri, [whitelist_entry])
            for entry in response.get('results', []):
                yield WhitelistEntry.fill_from_dict(entry)

        def delete_a_whitelist_entry(self, ip_address: str) -> dict:
            """Delete a whitelist entry

            url: https://docs.atlas.mongodb.com/reference/api/whitelist-delete-one/

            Args:
                ip_address (str): ip address to delete from whitelist

            Returns:
                dict: Response payload
            """
            uri = Settings.api_resources["Whitelist"]["Delete Whitelist Entry"] % (
                self.atlas.group, ip_address)
            return self.atlas.network.delete(Settings.BASE_URL + uri)

    class _MaintenanceWindows:
        """Maintenance Windows API

        see: https://docs.atlas.mongodb.com/reference/api/maintenance-windows/

        The maintenanceWindow resource provides access to retrieve or update the current Atlas project maintenance
         window. To learn more about Maintenance Windows, see the Set Preferred Cluster Maintenance Start Time setting
         on the View/Modify Project Settings page.

        Args:
            atlas (Atlas): Atlas instance
        """

        def __init__(self, atlas):
            self.atlas = atlas

        def _get_maint_window(self, as_obj: bool = True) -> Union[dict, MaintenanceWindow]:
            """
            (Internal)Gets the current maint window configuration for the the project.

            the current_config should be used instead.

            Args:
                as_obj: Return data as a MaintenanceWindowObj

            Returns:

            """
            uri = Settings.api_resources["Maintenance Windows"]["Get Maintenance Window"].format(
                GROUP_ID=self.atlas.group)
            response = self.atlas.network.get(Settings.BASE_URL + uri)
            if as_obj is False:
                return response
            else:
                return MaintenanceWindow.from_dict(response)

        def _update_maint_window(self, new_config: MaintenanceWindow) -> bool:
            """
            Uses the patch endpoint to update maint window settings.

            Args:
                as_obj: Return data as a MaintenanceWindowObj

            Returns:

            """

            uri = Settings.api_resources["Maintenance Windows"]["Update Maintenance Window"].format(
                GROUP_ID=self.atlas.group)
            self.atlas.network.patch(Settings.BASE_URL + uri, payload=new_config.as_update_dict())

            return True

        def _defer_maint_window(self) -> bool:
            """
            Used to defer the current maint window.

            TODO: Is this private method really needed? It is orignally here to provide a more flexible method to use
            if the public one was too simple, but may not be needed in the end.
            Returns:

            """
            uri = Settings.api_resources["Maintenance Windows"]["Defer Maintenance Window"].format(
                GROUP_ID=self.atlas.group)
            try:
                self.atlas.network.post(uri=Settings.BASE_URL + uri, payload={})
            except ErrAtlasBadRequest as e:
                if e.details.get('errorCode', None) == 'ATLAS_MAINTENANCE_NOT_SCHEDULED':
                    logger.warning(e.details.get('detail', 'No Detail available'))
                    return False
                else:
                    raise e
            return True

        def defer(self) -> dict:
            """
            Defers the currently scheduled maintenance window.

            Returns: bool:

            """
            output = self._defer_maint_window()
            return dict(maint_deffered=output)

        def current_config(self) -> MaintenanceWindow:
            """
            The current Maintainable Window configuration.

            Returns: MaintainableWindow object

            """
            return self._get_maint_window(as_obj=True)

        def set_config(self, new_config: MaintenanceWindow) -> bool:
            """
            Sets the maint configuration to the values in the passed MaintWindow Object

            Will only set those values which are not none in the MaintWindow Object. Currently you can not use
            this method to set a value as null. (This is not supported by the API anyway)

            Args:
                new_config: A MaintainenceWindow Object

            Returns: bool: True is success

            """
            output: bool = self._update_maint_window(new_config=new_config)
            return output

    class _CloudBackups:
        """Cloud Backup  API

        see: https://docs.atlas.mongodb.com/reference/api/cloud-backup/backup/backups/

        The CloudBackups resource provides access to retrieve the Cloud provider backup snapshots.

        Args:
            atlas (Atlas): Atlas instance
        """

        def __init__(self, atlas):
            self.atlas = atlas

        def create_snapshot_for_cluster(self, cluster_name: str, retention_days: int = 7,
                                        description: str = None, as_obj: bool = True) -> Union[
            CloudBackupSnapshot, dict]:
            """
            Creates and on demand snapshot for the passed cluster

            Args:
                as_obj:
                cluster_name:
                retention_days:
                description:
            """
            request_obj = CloudBackupRequest(cluster_name=cluster_name,
                                             retention_days=retention_days,
                                             description=description)

            uri = Settings.api_resources["Cloud Backup"]["Take an on-demand snapshot"] \
                .format(GROUP_ID=self.atlas.group, CLUSTER_NAME=cluster_name)

            try:
                response = self.atlas.network.post(uri=Settings.BASE_URL + uri, payload=request_obj.as_dict)
            except ErrAtlasBadRequest as e:
                logger.warning('Received an Atlas bad request on Snapshot creation. Could be due to overlap')
                raise IOError("Got a bad request error back from Atlas, this may be due to submitting"
                              "a snapshot request before a previous request has "
                              "completed.")
            logger.warning(f'Create response: {response}')
            if as_obj:
                logger.warning(f'Create response: {response}')
                return CloudBackupSnapshot(response)

            if not as_obj:
                return response

        def get_backup_snapshots_for_cluster(self, cluster_name: str, as_obj: bool = True) -> Union[
            Iterable[CloudBackupSnapshot], Iterable[dict]]:
            """Get  backup snapshots for a cluster.


            Retrieves
            url: https://docs.atlas.mongodb.com/reference/api/cloud-backup/backup/backups/

            Keyword Args:
                cluster_name (str) : The cluster name to fetch

            Returns:
                AtlasPagination or dict: Iterable object representing this function OR Response payload
            """

            uri = Settings.api_resources["Cloud Backup"]["Get all Cloud Backups for cluster"] \
                .format(GROUP_ID=self.atlas.group, CLUSTER_NAME=cluster_name)

            logger.debug(f"The cloudbackup uri used is {Settings.BASE_URL}{uri}")

            response = self.atlas.network.get(Settings.BASE_URL + uri)
            try:
                response_results: List[dict] = response['results']

            except KeyError:
                return_list = list()
                return_list.append(response)
                response_results: List[dict] = return_list

            if not as_obj:
                for entry in response_results:
                    yield entry
            if as_obj:
                for entry in response_results:
                    yield CloudBackupSnapshot.from_dict(entry)

        def get_backup_snapshot_for_cluster(self, cluster_name: str,
                                            snapshot_id: str, as_obj: bool = True) -> Union[
            Iterable[CloudBackupSnapshot], Iterable[dict]]:
            """Get  singe backup snapshot for a cluster.

            Retrieves
            url: https://docs.atlas.mongodb.com/reference/api/cloud-backup/backup/backups/

            Keyword Args:
                cluster_name (str) : The cluster name to fetch

            Returns:
                AtlasPagination or dict: Iterable object representing this function OR Response payload

            """
            logger.info('Getting by snapshotid')
            uri = Settings.api_resources["Cloud Backup"]["Get snapshot by SNAPSHOT-ID"] \
                .format(GROUP_ID=self.atlas.group, CLUSTER_NAME=cluster_name, SNAPSHOT_ID=snapshot_id)

            logger.debug(f"The cloudbackup uri used is {Settings.BASE_URL}{uri}")

            response = self.atlas.network.get(Settings.BASE_URL + uri)
            try:
                response_results: List[dict] = response['results']
            except KeyError:
                return_list = list()
                return_list.append(response)
                response_results: List[dict] = return_list

            if not as_obj:
                return list(response_results)

            if as_obj:
                for entry in response_results:
                    logger.debug(f'This is the cloud backup {entry}')
                    yield CloudBackupSnapshot.from_dict(entry)

        def is_existing_snapshot(self, cluster_name: str, snapshot_id: str) -> bool:
            try:
                out = self.atlas.CloudBackups.get_backup_snapshot_for_cluster(cluster_name, snapshot_id)
                for each in out:
                    assert each
                return True
            except (ErrAtlasNotFound, ErrAtlasBadRequest):
                return False

        def request_snapshot_restore(self, source_cluster_name: str, snapshot_id: str,
                                     target_cluster_name: str,
                                     delivery_type: DeliveryType = DeliveryType.automated,
                                     allow_same: bool = False
                                     ) -> SnapshotRestoreResponse:
            # Check if the target_cluster_name is valid
            if not self.atlas.Clusters.is_existing_cluster(target_cluster_name):
                logger.error(f'The passed target cluster {target_cluster_name}, does not exist in this project.')
                raise ValueError(f'The passed target cluster {target_cluster_name}, does not exist in this project.')
            else:
                logger.info('The target cluster exists.')
            # Check if the snapshot_id is valid
            if not self.atlas.CloudBackups.is_existing_snapshot(source_cluster_name, snapshot_id):
                error_text = f'The passed snapshot_id ({snapshot_id} ' \
                             f'is not valid for the source cluster {source_cluster_name})'
                logger.error(error_text)
                raise ValueError(error_text)
            else:
                logger.info('The snapshot_id is valid')

            # Check if the source and target clusters are the same, if so raise exception unless override is True

            if source_cluster_name == target_cluster_name and not allow_same:
                error_text = f'The source and target cluster should not be the same, as this is usually not the ' \
                             f'intended use case, and can lead to production data loss. If you wish to restore a ' \
                             f' snapshot to the same cluster, use the `allow_same` = True parameter.'
                logger.error(error_text)
                raise ValueError(error_text)

            uri = Settings.api_resources["Cloud Backup Restore Jobs"]["Restore snapshot by cluster"] \
                .format(GROUP_ID=self.atlas.group, CLUSTER_NAME=source_cluster_name)

            request_obj = SnapshotRestore(delivery_type, snapshot_id, target_cluster_name, self.atlas.group)

            try:
                response = self.atlas.network.post(uri=Settings.BASE_URL + uri, payload=request_obj.as_dict)
            except ErrAtlasBadRequest as e:
                if e.details.get('errorCode') == 'CLUSTER_RESTORE_IN_PROGRESS_CANNOT_UPDATE':
                    logger.error(e.details)
                    raise ErrAtlasRestoreConflictError(c=400, details=e.details)
                else:
                    logger.error('Received an Atlas bad request on Snapshot restore request.')
                    logger.error(e.details)
                    raise IOError("Received an Atlas bad request on Snapshot restore request.")

            try:
                response_obj = SnapshotRestoreResponse.from_dict(response)
            except KeyError as e:
                logger.error('Error encountered parsing response to a SnapshotRestoreResponse')
                logger.error(e)
                raise e
            return response_obj

        def request_snapshot_restore_to_group(self, source_cluster_name: str, snapshot_id: str,
                                              target_cluster_name: str,
                                              target_group_obj,
                                              delivery_type: DeliveryType = DeliveryType.automated

                                              ) -> SnapshotRestoreResponse:
            """Requests a snapshot restore to another group/project.

            Uses the passed target_group_obj, which is an Atlas object, to restore a snapshot from one group/project
            to another.

            This method does not check if the source and destination clusters have the same name, since this would
            not be dangerous when these are in two groups.

            Args:
                source_cluster_name: the text name of the source cluster
                snapshot_id: the uuid id of the snapshot to be restored
                target_cluster_name: the txt name of the destination cluster
                target_group_obj: Atlas: An Atlas object connected to the destination group.
                delivery_type: DeliveryType: IF you want to download, or automatically restore on Atlas.

            Returns:

            """
            # Check if the source and target groups are actually the same....
            if target_group_obj.group == self.atlas.group:
                txt = f"The source and target groups are the same ({self.atlas.group}). this method should only " \
                      f"be used if the restore is to anothe group than the target."
                logger.error(txt)
                raise AttributeError(txt)

            # Check if the target_cluster_name is valid
            if not target_group_obj.Clusters.is_existing_cluster(target_cluster_name):
                logger.error(f'The passed target cluster {target_cluster_name}, does not exist in this project.)')
                raise ValueError(f'The passed target cluster {target_cluster_name}, does not exist in this project.')
            else:
                logger.info('The target cluster exists in the other target group.')
            # Check if the snapshot_id is valid
            if not self.atlas.CloudBackups.is_existing_snapshot(source_cluster_name, snapshot_id):
                error_text = f'The passed snapshot_id ({snapshot_id} ' \
                             f'is not valid for the source cluster {source_cluster_name})'
                logger.error(error_text)
                raise ValueError(error_text)
            else:
                logger.info('The snapshot_id is valid')

            uri = Settings.api_resources["Cloud Backup Restore Jobs"]["Restore snapshot by cluster"] \
                .format(GROUP_ID=self.atlas.group, CLUSTER_NAME=source_cluster_name)

            request_obj = SnapshotRestore(delivery_type, snapshot_id, target_cluster_name, target_group_obj.group)

            try:
                response = self.atlas.network.post(uri=Settings.BASE_URL + uri, payload=request_obj.as_dict)
            except ErrAtlasBadRequest as e:
                if e.details.get('errorCode') == 'CLUSTER_RESTORE_IN_PROGRESS_CANNOT_UPDATE':
                    logger.error(e.details)
                    raise ErrAtlasRestoreConflictError(c=400, details=e.details)
                else:
                    logger.error('Received an Atlas bad request on Snapshot restore request.')
                    logger.error(e.details)
                    raise IOError("Received an Atlas bad request on Snapshot restore request.")

            try:
                response_obj = SnapshotRestoreResponse.from_dict(response)
            except KeyError as e:
                logger.error('Error encountered parsing response to a SnapshotRestoreResponse')
                logger.error(e)
                raise e
            return response_obj

        # Get all Cloud Backup restore jobs by cluster
        def get_snapshot_restore_requests(self, cluster_name: str, restore_id: str = None, as_obj: bool = True) \
                -> Union[List[Union[dict, SnapshotRestoreResponse]], SnapshotRestoreResponse, dict]:

            if not restore_id:
                uri = Settings.api_resources["Cloud Backup Restore Jobs"][
                    "Get all Cloud Backup restore jobs by cluster"] \
                    .format(GROUP_ID=self.atlas.group, CLUSTER_NAME=cluster_name)
            else:
                logger.info('Getting by restore_id')
                uri = Settings.api_resources["Cloud Backup Restore Jobs"]["Get Cloud Backup restore job by cluster"] \
                    .format(GROUP_ID=self.atlas.group, CLUSTER_NAME=cluster_name, JOB_ID=restore_id)

            logger.info(f"The restore job uri used is {Settings.BASE_URL}{uri}")

            response = self.atlas.network.get(Settings.BASE_URL + uri)

            try:
                response_results: List[dict] = response['results']
            except KeyError:
                return_list = list()
                return_list.append(response)
                response_results: List[dict] = return_list

            if not as_obj:
                if restore_id:
                    return list(response_results)
                else:
                    for entry in response_results:
                        yield entry
            if as_obj:
                for entry in response_results:
                    logger.debug(f'This is the {entry}')
                    yield SnapshotRestoreResponse.from_dict(entry)

        def cancel_snapshot_restore_request(self, cluster_name: str, restore_id: str):
            """
            Cancels a current backup restore request by restore_id.

            Calls:
            https://docs.atlas.mongodb.com/reference/api/cloud-backup/restore/delete-one-restore-job/

            Args:
                cluster_name: The name of the source cluster.
                restore_id:  The id of the (jobId) of the restore job.
            """
            # First Check if the restore_id is valid
            restore_info_result: SnapshotRestoreResponse = \
                self.get_snapshot_restore_requests(cluster_name=cluster_name, restore_id=restore_id, as_obj=True)
            restore_info = list(restore_info_result)[0]
            if not restore_info:
                raise ErrAtlasNotFound(404, {"error": f"The passed restore_id {restore_id} was not found"})

            elif restore_info.finished_at:
                raise ErrAtlasBadRequest(500,
                                         {"error": f"The passed restore_id ({restore_id} has already been completed"
                                                   f"and can not be canceled"})
            elif restore_info.cancelled is True:
                raise ErrAtlasConflict(500,
                                       {"error": f"The passed restore_id ({restore_id} has already been canceled"})

            else:
                uri = Settings.api_resources["Cloud Backup Restore Jobs"][
                    "Cancel manual download restore job by job_id"] \
                    .format(GROUP_ID=self.atlas.group, CLUSTER_NAME=cluster_name, JOB_ID=restore_id)
                logger.info(f'Preparing to cancel snapshot {restore_info.snapshot_id}, on'
                            f' {restore_info.target_cluster_name} with restore_id {restore_id}.')

                response = self.atlas.network.delete(Settings.BASE_URL + uri)

                return response

    class _Projects:
        """Atlas Projects

        see: https://www.mongodb.com/docs/atlas/reference/api/projects/

        The groups resource provides access to retrieve or create Atlas projects.

        Args:
            atlas (Atlas): Atlas instance
        """

        def __init__(self, atlas):
            self.atlas = atlas

        @property
        def projects(self) -> Iterable[Project]:
            """All Projects accessible by the current authed user/key
            Gets all projects for which the authed key has access.


            Returns (Iterable[Project]): Yields Project Objects.

            """

            uri = Settings.api_resources["Projects"]["Projects that the authenticated user can access"]

            try:
                response = self.atlas.network.get(uri=Settings.BASE_URL + uri)
            except Exception as e:
                raise e
            result_list = response["results"]

            for each in result_list:
                yield Project.from_dict(each)

        def _get_project(self, group_id: str = None, group_name: str = None) -> Project:
            """Returns a single Project
            Gets a single project, either by sending a group_id or a group name.

            Args:
                group_id (str): The unique identifier for the project.
                group_name (str): The user defined name for the project.

            Returns (Project): A Project Object.

            """
            if group_id and group_name:
                logger.error("Please pass either a group_id or a group_name, not both.")
                raise ValueError("Please pass either a group_id or a group_name, not both.")
            elif group_id:
                uri = Settings.api_resources["Projects"]["Project by group_id"].format(GROUP_ID=group_id)
            elif group_name:
                uri = Settings.api_resources["Projects"]["Project by group name"].format(GROUP_NAME=group_name)

            else:
                err_str = f"Please pass either a group_id or a group_name, you did not pass either. . ."
                logger.error(err_str)
                raise ValueError(err_str)

            try:
                response = self.atlas.network.get(uri=Settings.BASE_URL + uri)
            except Exception as e:
                raise e

            return Project.from_dict(response)

        def project_by_name(self, project_name: str) -> Project:
            """Return project by name

            Args:
                project_name (str): The project name to return

            Returns (Project): A single Project

            """
            return self._get_project(group_name=project_name)

        def project_by_id(self, project_id: str) -> Project:
            """Return project by name

            Args:
                project_id (str): The project id (group_id) to return

            Returns (Project): A single Project

            """
            return self._get_project(group_id=project_id)

        def _group_id_select(self, group_id: str = None) -> str:
            """Returns either the passed group_id or the instantiated group_id.

            Args:
                group_id (str):

            Returns (str): The correct group_id to use.
            """
            if not group_id:
                if not self.atlas.group:
                    raise ValueError(
                        "You either pass a group_id when calling get_project_teams, or instantiate the atlas"
                        "instance with a group_id, you have neither.")
                else:
                    group_id = self.atlas.group
            else:
                if self.atlas.group != group_id:
                    logger.warning(f"You have over-ridden the instantiated group_id {self.atlas.group} with the passed"
                                   f"group_id {group_id}. This is allowed but may yield unexpected results!")
                else:
                    logger.info("You have passed a override group which is the same as the instantiated atlast group,"
                                "this will have no effect")
            return group_id

        def get_project_teams(self, group_id: str = None) -> Iterable[TeamRoles]:
            """Retrieves all teams assigned to the passed project/group

            Returns each team assigned to the project, along with the roles which are assigned.


            Returns (Iterable[TeamRoles]): Yields TeamRole Objects.
            """
            group_id = self._group_id_select(group_id)

            uri = Settings.api_resources["Projects"]["Project teams by group_id"].format(GROUP_ID=group_id)

            try:
                response = self.atlas.network.get(uri=Settings.BASE_URL + uri)
            except Exception as e:
                raise e
            result_list = response["results"]

            for each in result_list:
                yield TeamRoles(each.get("teamId"), each.get("roleNames"))

        @staticmethod
        def _process_user_options(uri: str, flatten_teams: bool, include_org_users: bool) -> str:
            """Helper method to append user options to uri.

            Args:
                uri:
                flatten_teams:
                include_org_users:

            Returns (str): The processed uri string.
            """
            if flatten_teams and include_org_users:
                uri = uri + f"?flattenTeams={flatten_teams}&includeOrgUsers={include_org_users}"
            elif flatten_teams:
                uri = uri + f"?flattenTeams={flatten_teams}"
            elif include_org_users:
                uri = uri + f"?includeOrgUsers={include_org_users}"
            return uri

        def get_project_users(self, group_id: str = None, flatten_teams: Optional[bool] = None,
                              include_org_users: Optional[bool] = None) -> Iterable[AtlasUser]:
            """Yields all users (AtlasUser objects) associated with the group_id.

            Args:
                group_id (str): The group id to search, will use the configured group for the Atlas instance if instantiated in this way.
                flatten_teams (bool): Flag that indicates whether the returned list should include users who belong to a team that is assigned a role in this project. You might not have assigned the individual users a role in this project.
                include_org_users (bool): Flag that indicates whether the returned list should include users with implicit access to the project through the Organization Owner or Organization Read Only role. You might not have assigned the individual users a role in this project.


            Returns (Iterable[AtlasUser]: An iterable of AtlasUser objects.
            """
            group_id = self._group_id_select(group_id)
            uri = Settings.api_resources["Projects"]["Atlas Users assigned to project"].format(GROUP_ID=group_id)
            uri = self._process_user_options(uri, flatten_teams, include_org_users)

            try:
                response = self.atlas.network.get(uri=Settings.BASE_URL + uri)
            except Exception as e:
                raise e
            user_count: int = response["totalCount"]
            logger.error(f"The user count is {user_count}")
            result_list: List[dict] = response["results"]
            for each in result_list:
                yield AtlasUser.from_dict(each)

        def user_count(self, group_id: str = None, flatten_teams: Optional[bool] = None,
                       include_org_users: Optional[bool] = None,
                       ) -> int:
            """Returns count of users added to this project

            Args:
                group_id (str): The group id to search, will use the configured group for the Atlas instance if
                instantiated in this way.
                flatten_teams (bool): Flag that indicates whether the returned list should include users who belong to a
                 team that is assigned a role in this project. You might not have assigned the individual users a role
                 in this project.
                include_org_users (bool): Flag that indicates whether the returned list should include users with
                 implicit access to the project through the Organization Owner or Organization Read Only role. You
                  might not have assigned the individual users a role in this project.


            Returns (int): Count of users.

            """
            group_id = self._group_id_select(group_id)
            uri = Settings.api_resources["Projects"]["Atlas Users assigned to project"].format(GROUP_ID=group_id)
            uri = self._process_user_options(uri, flatten_teams, include_org_users)

            try:
                response = self.atlas.network.get(uri=Settings.BASE_URL + uri)
            except Exception as e:
                raise e
            user_count: int = response["totalCount"]
            logger.info(f"The user count is {user_count}")
            return user_count

        @property
        def settings(self) -> ProjectSettings:
            group_id = self.atlas.group
            uri = Settings.api_resources["Projects"]["Settings for project"].format(GROUP_ID=group_id)

            try:
                response = self.atlas.network.get(uri=Settings.BASE_URL + uri)
            except Exception as e:
                raise e
            return ProjectSettings.from_dict(response)

    class _Organizations:
        """Atlas Organizations

               see: https://www.mongodb.com/docs/atlas/reference/api/organizations/

               TThe orgs resource provides access to manage Atlas organizations.

               Args:
                   atlas (Atlas): Atlas instance
               """

        def __init__(self, atlas):
            self.atlas = atlas

        def _get_orgs(self, org_name: str = None, org_id: str = None) -> tuple:
            """Helper method for returning Org info from the API

            Args:
                org_name (str):  Organization name with which to filter the returned list. Performs a case-insensitive
                 search for organizations which exactly match the specified name.
                org_id (str): Specific org_id to return

            Returns (tuple): A tuple with results(array of dicts) and total count(int)

            """
            uri = Settings.api_resources["Organizations"]["Orgs the authenticated user can access"]
            if org_name:
                uri = uri + f"?name={org_name}"
            elif org_id:
                uri = uri + f"{org_id}"

            try:
                response = self.atlas.network.get(uri=Settings.BASE_URL + uri)
            except Exception as e:
                raise e

            if not org_id:
                result_list = response.get("results", [])
                total_count = response.get("TotalCount", 0)
            else:
                result_list = [response]
                total_count = 1

            return result_list, total_count

        @property
        def organizations(self) -> Iterable[Organization]:
            """All Organizations accessible by the current authed user/key
            Gets all Organizations for which the authed key has access.


            Returns (Iterable[Organization]): Yields Organization Objects.

            """
            result_list = self._get_orgs()[0]
            for each in result_list:
                yield Organization.from_dict(each)

        def organization_by_name(self, org_name: str) -> Organization:
            """Single organization searched by name.

            Args:
                org_name: Organization name with which to filter the returned list. Performs a case-insensitive
                 search for organizations which exactly match the specified name.

            Returns (Organization): a single Organization object.

            """
            result = self._get_orgs(org_name=org_name)[0]
            if isinstance(result, list):
                return Organization.from_dict(result[0])
            else:
                return Organization.from_dict(result)

        def organization_by_id(self, org_id: str) -> Organization:
            """Single organization searched by org_id.

            Args:
                org_id (str):
            Returns (Organization): a single Organization object.

            """
            result = self._get_orgs(org_id=org_id)
            if isinstance(result[0], list):
                logger.error("returning a list?")
                return Organization.from_dict(result[0][0])
            else:
                return Organization.from_dict(result[0])

        @property
        def count(self) -> int:
            """ Count of Organizations accessible by the authed user/key.

            Returns (int):

            """
            return self._get_orgs()[1]

        def get_all_projects_for_org(self, org_id: str) -> Iterable[Project]:
            """Get projects related to the current Org

            url: https://www.mongodb.com/docs/atlas/reference/api/organization-get-all-projects/

            Args:
                org_id (str): The organization id which owns the projects.

            Returns:
                Iterable[Project]: Iterable containing projects


            """
            uri: str = Settings.api_resources["Organizations"]["Projects associated with the Org"].format(
                ORG_ID=org_id)

            response = self.atlas.network.get_big(Settings.BASE_URL + uri)
            for entry in response.get('results', []):
                yield Project.from_dict(entry)


class AtlasPagination:
    """Atlas Pagination Generic Implementation

    Constructor

    Args:
        atlas (Atlas): Atlas instance
        fetch (function): The function "get_all" to call
        pageNum (int): Page number
        itemsPerPage (int): Number of Users per Page
    """

    def __init__(self, atlas, fetch, pageNum: int, itemsPerPage: int):
        self.atlas = atlas
        self.fetch = fetch
        self.pageNum = pageNum
        self.itemsPerPage = itemsPerPage

    def __iter__(self):
        """Iterable

        Yields:
            str: One result
        """

        # pageNum is set with the value requested (so not necessary 1)
        pageNum = self.pageNum
        # total: This is a fake value to enter into the while. It will be updated with a real value later
        total = pageNum * self.itemsPerPage

        while pageNum * self.itemsPerPage - total < self.itemsPerPage:
            # fetch the API
            try:
                details = self.fetch(pageNum, self.itemsPerPage)
            except Exception as e:
                raise ErrPagination(e)

            # set the real total
            total = details["totalCount"]

            # while into the page results
            results = details["results"]
            results_count = len(results)
            index = 0
            while index < results_count:
                result = results[index]
                index += 1
                yield result

            class DatabaseUsersGetAll(AtlasPagination):
                """Pagination for Database User : Get All"""

                def __init__(self, atlas, pageNum, itemsPerPage):
                    super().__init__(atlas, atlas.DatabaseUsers.get_all_database_users, pageNum, itemsPerPage)

            # next page
            pageNum += 1


class ClustersGetAll(AtlasPagination):
    """Pagination for Clusters : Get All"""

    def __init__(self, atlas, pageNum, itemsPerPage):
        super().__init__(atlas, atlas.Clusters.get_all_clusters, pageNum, itemsPerPage)


# noinspection PyProtectedMember
class HostsGetAll(AtlasPagination):
    """Pagination for Processes : Get All"""

    def __init__(self, atlas: Atlas, pageNum: int, itemsPerPage: int):
        super().__init__(atlas, atlas.Hosts._get_all_hosts, pageNum, itemsPerPage)


# noinspection PyProtectedMember
class EventsGetForProject(AtlasPagination):

    def __init__(self, atlas: Atlas, since_datetime: datetime, pageNum: int, itemsPerPage: int):
        super().__init__(atlas, self.fetch, pageNum, itemsPerPage)
        self._get_all_project_events = atlas.Events._get_all_project_events
        self.since_datetime = since_datetime

    def fetch(self, pageNum, itemsPerPage):
        """Intermediate fetching

        Args:
            pageNum (int): Page number
            itemsPerPage (int): Number of Events per Page

        Returns:
            dict: Response payload
        """
        return self._get_all_project_events(self.since_datetime, pageNum, itemsPerPage)


class DatabaseUsersGetAll(AtlasPagination):
    """Pagination for Database User : Get All"""

    def __init__(self, atlas, pageNum, itemsPerPage):
        super().__init__(atlas, atlas.DatabaseUsers.get_all_database_users, pageNum, itemsPerPage)


class AlertsGetAll(AtlasPagination):
    """Pagination for Alerts : Get All"""

    def __init__(self, atlas, status, pageNum, itemsPerPage):
        super().__init__(atlas, self.fetch, pageNum, itemsPerPage)
        self.get_all_alerts = atlas.Alerts.get_all_alerts
        self.status = status

    def fetch(self, pageNum, itemsPerPage):
        """Intermediate fetching

        Args:
            pageNum (int): Page number
            itemsPerPage (int): Number of Users per Page

        Returns:
            dict: Response payload
        """
        return self.get_all_alerts(self.status, pageNum, itemsPerPage)


class WhitelistGetAll(AtlasPagination):
    """Pagination for Database User : Get All"""

    def __init__(self, atlas, pageNum, itemsPerPage):
        super().__init__(atlas, atlas.Whitelist.get_all_whitelist_entries, pageNum, itemsPerPage)


class CloudBackupSnapshotsGetAll(AtlasPagination):
    """Pagination for Database User : Get All"""

    def __init__(self, atlas, pageNum, itemsPerPage):
        super().__init__(atlas, atlas.CloudBackups.get_all_backup_snapshots, pageNum, itemsPerPage)


class OrganizationProjectsGetAll(AtlasPagination):
    """Pagination for Database User : Get All"""

    def __init__(self, atlas: Atlas, org_id: str, pageNum, itemsPerPage):
        super().__init__(atlas, self.fetch, pageNum, itemsPerPage)
        print(f"Got to init")
        self._get_all_projects_for_org = atlas.Organizations.get_all_projects_for_org
        self.org_id = org_id

    def fetch(self, pageNum, itemsPerPage):
        """Intermediate fetching

        Args:
            pageNum (int): Page number
            itemsPerPage (int): Number of Events per Page

        Returns:
            dict: Response payload
        """
        print("Got to fetch")
        return self._get_all_projects_for_org(self.org_id, pageNum, itemsPerPage)



# Copyright (c) 2021 Matthew G. Monteleone
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Cloud Backups Module

Provides access to Cloud Backups and Cloud backup restore endpoints
"""

from enum import Enum
from typing import List, NewType, Optional
from datetime import datetime
from atlasapi.lib import ProviderName, ClusterType
from dateutil.parser import parse
from pprint import pprint
from distutils.util import strtobool
import logging

logger = logging.getLogger(name='Atlas_cloud_backup')


def try_date(str_in: str) -> Optional[datetime]:
    try:
        datetime_out = parse(str_in)
    except (ValueError, TypeError, AttributeError):
        logger.debug(f'Could not parse a date from : {str_in}')
        datetime_out = None
    return datetime_out


def try_bool(str_in: str) -> bool:
    if type(str_in) != bool:
        try:
            bool_out = strtobool(str_in)
        except (ValueError, TypeError, AttributeError):
            logger.debug(f'Could not parse a bool from : {str_in}')
            bool_out = False
    else:
        bool_out = str_in
    return bool_out


class SnapshotType(Enum):
    ONDEMAND = "On Demand"
    SCHEDULED = "Scheduled"
    FALLBACK = "Fallback"


class SnapshotStatus(Enum):
    QUEUED = "Queued"
    INPROGRESS = "In Progress"
    COMPLETED = "Completed"
    FAILED = "Failed"


class DeliveryType(Enum):
    automated = "Automated restore to Atlas cluster"
    download = "manual download of archived data directory"
    pointInTime = "Automated point in time restore to Atlas Cluster"


class SnapshotRestore(object):
    def __init__(self, delivery_type: DeliveryType,
                 snapshot_id: str,
                 target_cluster_name: str = None,
                 target_group_id: str = None):
        self.deliveryType = delivery_type
        self.snapshot_id = snapshot_id
        self.target_cluster_name = target_cluster_name
        self.target_group_id = target_group_id

    @property
    def as_dict(self) -> dict:
        return dict(
            deliveryType=self.deliveryType.name,
            snapshotId=self.snapshot_id,
            targetClusterName=self.target_cluster_name,
            targetGroupId=self.target_group_id
        )


class SnapshotRestoreResponse(SnapshotRestore):
    def __init__(self, restore_id: str, delivery_type: DeliveryType, snapshot_id: str, target_cluster_name: str,
                 target_group_id: str, cancelled: bool = False,
                 created_at: datetime = None, expired: bool = False, expires_at: datetime = None,
                 finished_at: datetime = None, links: list = None, snapshot_timestamp: datetime = None,
                 target_deployment_item_name: str = None,
                 delivery_url: str = None):
        super().__init__(delivery_type, snapshot_id, target_cluster_name, target_group_id)
        self.delivery_url = delivery_url
        self.target_deployment_item_name = target_deployment_item_name
        self.snapshot_timestamp = snapshot_timestamp
        self.links = links
        self.finished_at = finished_at
        self.expires_at = expires_at
        self.expired = expired
        self.created_at = created_at
        self.cancelled = cancelled
        self.restore_id = restore_id

    @classmethod
    def from_dict(cls, data_dict):
        restore_id = data_dict.get('id')
        snapshot_id = data_dict.get('snapshotId')
        try:
            delivery_type = DeliveryType[data_dict.get('deliveryType')]
        except KeyError:
            logger.warning(f'Got an unmapped deliveryType : {data_dict.get("deliveryType")}.')
            delivery_type = None
        target_cluster_name = data_dict.get('targetClusterName')
        target_group_id = data_dict.get('targetGroupId')
        cancelled = try_bool(data_dict.get('cancelled'))
        created_at = try_date(data_dict.get('createdAt')) # Does not appear to be retunred at all, potential docs issue?
        expired = try_bool(data_dict.get('expired'))
        expires_at = try_date(data_dict.get('expiresAt')) # Does not appear to be retunred at all, potential docs issue?
        finished_at = try_date(data_dict.get('finishedAt')) # Does not appear to be retunred at all, potential docs issue?
        links = data_dict.get('links')
        snapshot_timestamp = try_date(data_dict.get('timestamp'))
        target_deployment_item_name=data_dict.get('targetDeploymentItemName') # missing in documentation TODO: File docs ticket.
        delivery_url = data_dict.get('deliveryUrl',None) # missing in documentation TODO: File docs ticket

        return cls(restore_id=restore_id,
                   delivery_type=delivery_type,
                   snapshot_id=snapshot_id,
                   target_cluster_name=target_cluster_name,
                   target_group_id=target_group_id,
                   cancelled=cancelled,
                   created_at=created_at,
                   expired=expired,
                   expires_at=expires_at,
                   finished_at=finished_at,
                   links=links,
                   snapshot_timestamp=snapshot_timestamp,
                   target_deployment_item_name=target_deployment_item_name,
                   delivery_url=delivery_url)


class CloudBackupRequest(object):
    def __init__(self, cluster_name: str, retention_days: int = 1, description: str = 'Created by pyAtlasAPI') -> None:
        self.description = description
        self.retentionInDays = retention_days
        self.cluster_name = cluster_name

    @property
    def as_dict(self):
        return dict(description=self.description, retentionInDays=self.retentionInDays)


class CloudBackupSnapshot(object):
    def __init__(self, id: Optional[str] = None,
                 cloud_provider: Optional[ProviderName] = None,
                 created_at: Optional[datetime] = None,
                 description: Optional[str] = None,
                 expires_at: Optional[datetime] = None,
                 links: Optional[List] = None,
                 masterkey_uuid: Optional[str] = None,
                 members: Optional[list] = None,
                 mongod_version: Optional[str] = None,
                 replica_set_name: Optional[str] = None,
                 snapshot_ids: Optional[list] = None,
                 snapshot_type: Optional[SnapshotType] = None,
                 status: Optional[SnapshotStatus] = None,
                 storage_size_bytes: Optional[int] = None,
                 type: Optional[ClusterType] = None):
        """
        Details of a Cloud Provider Snapshot.

        Args:
            id: Unique identifier of the snapshot.
            cloud_provider: Cloud provider that stores this snapshot. Atlas returns this parameter
            when "type": "replicaSet".
            created_at:
            description: Description of the snapshot. Atlas returns this parameter when "status": "onDemand".
            expires_at:
            links: One or more links to sub-resources and/or related resources. The relations between URLs are
                   explained in the Web Linking Specification
            masterkey_uuid: Unique identifier of the AWS KMS Customer Master Key used to encrypt the snapshot.
                            Atlas returns this value for clusters using Encryption at Rest via Customer KMS.
            members: List of snapshots and the cloud provider where the snapshots are stored.
                     Atlas returns this parameter when "type": "shardedCluster".
            mongod_version:
            replica_set_name: Label given to the replica set from which Atlas took this snapshot.
                              Atlas returns this parameter when "type": "replicaSet".
                              snapshot_ids: Unique identifiers of the snapshots created for the shards and config server
                              for a sharded cluster. Atlas returns this parameter when "type": "shardedCluster".
                              These identifiers should match those given in the members[n].id parameters.
                              This allows you to map a snapshot to its shard or config server name.
            snapshot_type: Type of snapshot. Atlas can return onDemand or scheduled.
            status: Current status of the snapshot. Atlas can return one of the following values:
                    (queued, inProgress, completed, failed)
            storage_size_bytes:
            type: Type of cluster. Atlas can return replicaSet or shardedCluster.


        """
        self.type: Optional[ClusterType] = type
        self.storage_size_bytes: Optional[int] = storage_size_bytes
        self.status: Optional[SnapshotStatus] = status
        self.snapshot_type: Optional[SnapshotType] = snapshot_type
        self.snapshot_ids: Optional[list] = snapshot_ids
        self.replica_set_name: Optional[str] = replica_set_name
        self.mongod_version: Optional[str] = mongod_version
        self.members: Optional[list] = members
        self.masterkey_uuid: Optional[str] = masterkey_uuid
        self.links: Optional[list] = links
        self.expires_at: Optional[datetime] = expires_at
        self.description: Optional[str] = description
        self.created_at: Optional[datetime] = created_at
        self.cloud_provider: Optional[ProviderName] = cloud_provider
        self.id: Optional[str] = id

    @classmethod
    def from_dict(cls, data_dict: dict):
        id = data_dict.get('id', None)
        try:
            cloud_provider = ProviderName[data_dict.get('cloudProvider')]
        except KeyError:
            cloud_provider = ProviderName.TENANT
        created_at = try_date(data_dict.get('createdAt'))
        expires_at = try_date(data_dict.get('expiresAt'))
        description = data_dict.get('description')
        snapshot_type = SnapshotType[data_dict.get('snapshotType').upper()]
        cluster_type = ClusterType[data_dict.get('type').upper()]
        snapshot_status = SnapshotStatus[data_dict.get('status').upper()]
        storage_size_bytes = data_dict.get('storageSizeBytes')
        replica_set_name = data_dict.get('replicaSetName')
        links = data_dict.get('links')
        masterkey = data_dict.get('masterKeyUUID')
        members = data_dict.get('members')
        mongod_version = data_dict.get('mongodVersion')
        snapshot_ids = data_dict.get('snapshotIds')
        return cls(id=id,
                   cloud_provider=cloud_provider,
                   created_at=created_at,
                   expires_at=expires_at,
                   description=description,
                   snapshot_type=snapshot_type,
                   type=cluster_type,
                   status=snapshot_status,
                   storage_size_bytes=storage_size_bytes,
                   replica_set_name=replica_set_name,
                   links=links,
                   masterkey_uuid=masterkey,
                   members=members,
                   mongod_version=mongod_version,
                   snapshot_ids=snapshot_ids
                   )



# Copyright (c) 2019 Matthew G. Monteleone
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import NewType, List, Optional, Union, Dict
from dateutil.parser import parse
import logging
from datetime import datetime

from atlasapi.events_event_types import AtlasEventTypes
import ipaddress
from copy import copy

logger = logging.getLogger(name='Atlas_events')


class _AtlasBaseEvent(object):
    def __init__(self, value_dict: dict) -> None:
        self.created_date = None
        try:
            self.created_date = parse(value_dict.get("created", None))
        except ValueError as e:
            logger.warning("Could not parse datetime value for created_date: {}".format(e))
            pass
        self.event_type = AtlasEventTypes[value_dict.get('eventTypeName', 'UNKNOWN')]  # type: AtlasEventTypes
        self.group_id = value_dict.get('groupId', None)  # type: str
        self.id = value_dict.get('id', None)  # type: str
        self.is_global_admin = value_dict.get('isGlobalAdmin', False)  # type: bool
        self.links = value_dict.get('links', None)  # type: list
        self.event_dict = value_dict  # type: dict
        self.additional_data = value_dict.get('raw', None)

    def as_dict(self):
        original_dict = self.__dict__
        return_dict = copy(original_dict)
        del return_dict['event_dict']
        del return_dict['additional_data']
        return_dict['created_date'] = datetime.isoformat(self.created_date)
        return_dict['event_type'] = self.event_type.name
        return_dict['event_type_desc'] = self.event_type.value

        if return_dict.get('remote_address'):
            return_dict['remote_address'] = return_dict['remote_address'].__str__()
        return return_dict


class _AtlasUserBaseEvent(_AtlasBaseEvent):
    def __init__(self, value_dict: dict) -> None:
        super().__init__(value_dict)
        self.user_id = value_dict.get('userId', None)  # type: str
        self.username = value_dict.get('username')  # type: str
        try:
            self.remote_address = ipaddress.ip_address(value_dict.get('remoteAddress', None))  # type: ipaddress
        except ValueError:
            logger.info('No IP address found')
            self.remote_address = None


class AtlasEvent(_AtlasBaseEvent):
    def __init__(self, value_dict: dict) -> None:
        super().__init__(value_dict)


class AtlasCPSEvent(_AtlasBaseEvent):
    def __init__(self, value_dict: dict) -> None:
        """
        Atlas Events for Cloud Provider Snapshot related events.

        Contains extra data points directly related to CPS events. There
        !NOTE! The extra data points are extracted from the "raw" property which is not guaranteed to be stable.

        Args:
            value_dict: The original dict of values from the Atlas API
        """
        super().__init__(value_dict)
        if self.additional_data:
            self.snapshot_id: Optional[str] = self.additional_data.get('snapshotId', None)
            self.snapshot_completion_date: Optional[datetime] = None
            self.snapshot_scheduled_creation_date: Optional[datetime] = None
            self.cluster_name: str = self.additional_data.get('clusterName', None)
            self.cluster_id: str = self.additional_data.get('clusterId', None)
            snapshot_completion_date = None
            snapshot_scheduled_creation_date = None
            try:
                snapshot_completion_date = self.additional_data.get('snapshotCompletionDate', None)
                self.snapshot_completion_date: Optional[datetime] = parse(snapshot_completion_date)
            except (ValueError, TypeError, AttributeError):
                logger.debug(f'Could not parse a CPS snapshot completion date: {snapshot_completion_date}')
            try:
                snapshot_scheduled_creation_date = self.additional_data.get('snapshotScheduledCreationDate', None)
                self.snapshot_scheduled_creation_date: Optional[datetime] = parse(snapshot_scheduled_creation_date)
            except (ValueError, TypeError, AttributeError):
                logger.debug(
                    f'Could not parse a CPS snapshot scheduled creation date: {snapshot_scheduled_creation_date}')


class AtlasDataExplorerEvent(_AtlasUserBaseEvent):
    def __init__(self, value_dict: dict) -> None:
        super().__init__(value_dict)
        self.database = value_dict.get('database', None)  # type: str
        self.collection = value_dict.get('collection', None)  # type: str
        self.op_type = value_dict.get('opType', None)  # type: str


class AtlasClusterEvent(_AtlasBaseEvent):
    def __init__(self, value_dict: dict) -> None:
        super().__init__(value_dict)
        self.replica_set_name = value_dict.get('replicaSetName', None)  # type: str
        self.cluster_name = value_dict.get('clusterName', None)  # type: str


class AtlasHostEvent(_AtlasBaseEvent):
    def __init__(self, value_dict: dict) -> None:
        super().__init__(value_dict)
        self.hostname = value_dict.get('hostname', None)  # type: str
        self.port = value_dict.get('port', None)  # type: int
        self.replica_set_name = value_dict.get('replicaSetName', None)  # type: str


class AtlasFeatureEvent(_AtlasUserBaseEvent):
    def __init__(self, value_dict: dict) -> None:
        super().__init__(value_dict)
        self.hostname = value_dict.get('hostname', None)  # type: str
        self.feature_name = value_dict.get('featureName', None)  # type: str


def atlas_event_factory(value_dict: dict) -> Union[
    AtlasEvent, AtlasDataExplorerEvent, AtlasClusterEvent, AtlasHostEvent, AtlasFeatureEvent, AtlasCPSEvent]:
    if 'CPS_' in value_dict.get("eventTypeName", None):
        return AtlasCPSEvent(value_dict=value_dict)
    elif value_dict.get("featureName", None):
        return AtlasFeatureEvent(value_dict=value_dict)
    elif value_dict.get("hostname", None):
        return AtlasHostEvent(value_dict=value_dict)

    elif value_dict.get("clusterName", None):
        return AtlasClusterEvent(value_dict=value_dict)

    elif value_dict.get("database", None):
        return AtlasDataExplorerEvent(value_dict=value_dict)


    else:
        return AtlasEvent(value_dict=value_dict)


ListOfEvents = NewType('ListOfEvents', List[Union[Dict, _AtlasBaseEvent]])



# Copyright (c) 2019 Joe Drumgoole
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modifications copyright (C) Joe Drumgoole

from enum import Enum
import pprint


class ListFormat(Enum):
    short = "short"
    full = "full"

    def __str__(self):
        return self.value


class Commands(Enum):
    List = "List"

    def __str__(self):
        return self.value


class ListCommand:
    """
    Execute the list command
    """

    def __init__(self, format: ListFormat = ListFormat.short):
        self._format = format

    def list_one(self, resource):
        if self._format is ListFormat.short:
            print(resource["id"])
        else:
            pprint.pprint(resource)

    def list_all(self, iterator):
        counter = 0
        for counter, resource in enumerate(iterator):
            self.list_one(resource)
        return counter



# Copyright (c) 2019 Joe Drumgoole
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modifications copyright (C) Joe Drumgoole

from enum import Enum
import os
from atlascli.atlaserrors import AtlasEnvironmentError


class AtlasEnv(Enum):
    ATLAS_PUBLIC_KEY = "ATLAS_PUBLIC_KEY"
    ATLAS_PRIVATE_KEY = "ATLAS_PRIVATE_KEY"

    def __str__(self):
        return self.value


class AtlasKey:

    def __init__(self, public_key, private_key):

        self._public_key = public_key
        self._private_key = private_key

    @staticmethod
    def getenv(key_string):
        key = os.getenv(key_string)
        if key is None:
            raise AtlasEnvironmentError(f"Private key environment variable '{key_string}' is not set")
        return key

    @classmethod
    def get_from_env(cls):
        public_key = AtlasKey.getenv(AtlasEnv.ATLAS_PUBLIC_KEY.value)
        private_key = AtlasKey.getenv(AtlasEnv.ATLAS_PRIVATE_KEY.value)
        return AtlasKey(public_key, private_key)

    @property
    def private_key(self):
        return self._private_key

    @property
    def public_key(self):
        return self._public_key

    @staticmethod
    def obfuscate(s, show=4, hide_char="x"):
        l = len(s)
        if show > l:
            return s
        else:
            return (hide_char * (l - show)) + s[:show]

    def __repr__(self):
        return (f"AtlasKey(public_key='{AtlasKey.obfuscate(self._public_key)}', " +
                f"private_key='{AtlasKey.obfuscate(self._private_key)}')")



# Copyright (c) 2019 Joe Drumgoole
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modifications copyright (C) Joe Drumgoole

import argparse
import sys
# import pprint
import logging
import os
from enum import Enum

from atlasapi.atlas import Atlas

from atlascli.listcommand import ListCommand, ListFormat


class AtlasResource(Enum):
    ORGANIZATION = "organization"
    PROJECT = "project"
    CLUSTER = "cluster"

    def __str__(self):
        return self.value


def main(args):
    """

    :param args: Expect sys.argv
    :return: None
    """
    parser = argparse.ArgumentParser(prog="atlascli",
                                     description="A command line interface to the MongoDB Atlas"
                                                 "database as a service."
                                                 "https://www.mongodb.com/cloud/atlas for more info"
                                                 "See also https://docs.atlas.mongodb.com/configure-api-access"
                                                 "/#programmatic-api-keys "
                                                 "For how to obtain a programmatic API key required to access the API"
                                     )

    parser.add_argument("--publickey", help="MongoDB Atlas public API key")
    parser.add_argument("--privatekey", help="MongoDB Atlas private API key")
    parser.add_argument("--atlasgroup", help="Default group (aka project)")

    parser.add_argument("--format", type=ListFormat,
                        choices=list(ListFormat),
                        default=ListFormat.short,
                        help="Format for output of list command [default: %(default)s]")

    parser.add_argument("--resource",
                        type=AtlasResource, default=AtlasResource.CLUSTER,
                        choices=list(AtlasResource),
                        help="Which resource type are we operating on:"
                             "organization, project or cluster? [default: %(default)s]")

    parser.add_argument('--id', type=str, help='Specify a resource id')

    parser.add_argument("--debug", default=False, action="store_true",
                        help="Turn on logging at debug level [default: %(default)s]")

    parser.add_argument("--list", default=False, action="store_true",
                        help="List a set of resources [default: %(default)s]")

    args = parser.parse_args(args)

    if args.debug:
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                            level=logging.DEBUG)
    else:
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                            level=logging.INFO)

    logging.debug("logging is on at DEBUG level")

    if args.publickey:
        public_key = args.publickey
    else:
        public_key = os.getenv("ATLAS_PUBLIC_KEY")
        if public_key is None:
            print("you must specify an ATLAS public key via --publickey arg "
                  "or the environment variable ATLAS_PUBLIC_KEY")
            sys.exit(1)

    if args.privatekey:
        private_key = args.privatekey
    else:
        private_key = os.getenv("ATLAS_PRIVATE_KEY")
        if private_key is None:
            print("you must specify an an ATLAS private key via --privatekey"
                  "arg or the environment variable ATLAS_PRIVATE_KEY")
            sys.exit(1)

    atlas = Atlas(public_key, private_key, args.atlasgroup)

    if args.list:
        if args.resource == AtlasResource.CLUSTER:
            list_cmd = ListCommand(args.format)
            if args.id:
                print("Cluster:")
                cluster = atlas.Clusters.get_single_cluster(cluster=args.id)
                list_cmd.list_one(cluster)
            else:
                print("Cluster list:")
                clusters = atlas.Clusters.get_all_clusters(iterable=True)
                total = list_cmd.list_all(clusters)
                print(f"{total} cluster(s)")


if __name__ == "__main__":
    main(sys.argv[1:])  # strip off the program name



from typing import Optional, NewType, List
from datetime import datetime

OptionalInt = NewType('OptionalInt', Optional[int])
OptionalStr = NewType('OptionalStr', Optional[str])
OptionalDateTime = NewType('OptionalDateTime', Optional[datetime])
OptionalBool = NewType('OptionalBool', Optional[bool])
ListOfStr = NewType('ListofStr', List[str])
ListofDict = NewType('ListOfDict', List[str])
OptionalFloat = NewType('OptionalFloat', Optional[float])